# vsender.pm
# ==========
# [Spamassassin Plugin that check if sender recipient is able to accept mail]


package Mail::SpamAssassin::Plugin::vsender;

use strict;
use Mail::SpamAssassin;
use Mail::SpamAssassin::Plugin;
use Mail::SpamAssassin::Logger;
use Mail::SpamAssassin::Timeout;

use Net::DNS;
use Net::SMTP;
use DBI;
use vars qw(@ISA);


# Environment
my $driver   = "SQLite";
my $database = "/etc/mail/spamassassin/vsender/vsender.db";   # Full database path
my $dsn = "DBI:$driver:dbname=$database";
our $HELO = 'yopurdomain.net';                                # Helo string for testing, mailserver name
our $CACHE_DAYS_OK = 30;                                      # Days to mantain in cache correctly verified from addresses
our $CACHE_DAYS_KO = 10;                                      # Days to mantain in cache from addresses that was unable to verify
our $CACHE_DAYS_GREY = 50;                                    # Days to mantain in cache domains that use greylist

@ISA = qw(Mail::SpamAssassin::Plugin);

my $VERSION = "1.0";

sub new {
  my ($class, $mailsa) = @_;
  $class = ref($class) || $class;
  my $self = $class->SUPER::new($mailsa);
  bless ($self, $class);
  # Not for local tests
  if ($mailsa->{local_tests_only}) {
    dbg("vsender: local tests only, vsender off"); }
  else {
    dbg("vsender: Plugin vsender $VERSION");
  }
  $self->register_eval_rule ("vs_check");
  $self->register_eval_rule ("vs_grey");
  return $self;
}

sub vs_grey {
  my ($self, $permsgstatus) = @_;
  my $smtp;
  my $dbh;

  # Exit for local tests
  return 0 if $self->{main}->{local_tests_only}; 

  # Exit if no dns available
  return 0 unless $permsgstatus->is_dns_available();

  # Exit for all_trusted ip
  return 0 if $permsgstatus->check_all_trusted;

  # Connect database
  if( ! -f $database ){
    dbg("vsender_grey: Missing database, exiting");
    return -1;
  } else {
    $dbh = DBI->connect($dsn, "", "", { RaiseError => 1 }) or dbg("vsender_grey: Database error: " .  $DBI::errstr );
  }

  # From scan loop
  foreach my $from ($permsgstatus->all_from_addrs) {
   dbg("vsender_grey: testing from: $from");

   # Splitting domain of from user   
   my( $user, $domain ) = split( "@", $from );
   dbg("vsender_grey: testing domain: $domain");
   
   # Search 'domain' in db cache
   my $result =  $dbh->selectrow_hashref( "SELECT * FROM vsendergrey WHERE vdomain = ?", {}, ( $domain ) ) ;
   if( $result ) {
    dbg("vsender_grey: found $domain cached in database, this is a greylisting domain");
    return 1;
   } else {
    dbg("vsender_grey: not found $domain cached in database, this is not a greylisting domain");
    return 0;
   } 
  }
}

sub vs_check {
  my ($self, $permsgstatus) = @_;
  my $smtp;
  my $dbh;

  # Exit for local tests
  return 0 if $self->{main}->{local_tests_only}; 

  # Exit if no dns available
  return 0 unless $permsgstatus->is_dns_available();

  # Exit for all_trusted ip
  return 0 if $permsgstatus->check_all_trusted;

  # Connect database
  if( ! -f $database ){
    dbg("vsender_check: Missing database, exiting");
    return -1;
  } else {
    $dbh = DBI->connect($dsn, "", "", { RaiseError => 1 }) or dbg("vsender_check: Database error: " .  $DBI::errstr );
  }

  # Db Manteinance
  # 0 means that was unable to verify, 1 means that was correctly verified
   $dbh->do("DELETE FROM vsender WHERE timestamp < date( 'now','-$CACHE_DAYS_OK day');") or dbg("vsender_check: Database error: " .  $DBI::errstr );
   $dbh->do("DELETE FROM vsender WHERE timestamp < date( 'now','-$CACHE_DAYS_KO day');") or dbg("vsender_check: Database error: " .  $DBI::errstr );
   $dbh->do("DELETE FROM vsendergrey WHERE timestamp < date( 'now','-$CACHE_DAYS_GREY day');") or dbg("vsender_check: Database error: " .  $DBI::errstr );

  # Find the latest 'to' of this email to use for testing as 'from'
  my $totest = 'nobody@test.it';
  foreach my $to ($permsgstatus->all_to_addrs) { $totest = $to; }
  dbg("vsender_check: this 'to' will be used to test sender: $totest");

  # From scan loop
  foreach my $from ($permsgstatus->all_from_addrs) {
   dbg("vsender_check: testing from: $from");
   
   # Splitting domain of from user   
   my( $user, $domain ) = split( "@", $from );
   dbg("vsender_check: testing domain: $domain");

   # Exit if null domain
   if ( ! $domain ){ dbg("vsender_check: Couldn't find sender domain in permsgstatus->all_from_addrs"); return 0; }

   # Search 'from' in db cache
   my $result =  $dbh->selectrow_hashref( "SELECT * FROM vsender WHERE vfrom = ?", {}, ( $from ) ) ;
   if( $result ) {
    if( $result->{vok} ) {
     dbg("vsender_check: found $from cached in database, and VERIFIED");
     return 1;
    } else {
     dbg("vsender_check: found $from cached in database, and NOT VERIFIED");
     return 0;
    }
   } 

   # If we are here this is not cached compose possible list of mailserver 
   # a list of all mx records server and domain itself that ican be used
   # when mx not defined
   my @mxlist;
   my $res = Net::DNS::Resolver->new;
   my $packet = $res->query( $domain, 'MX' );
   if ( $packet ) {
    foreach my $answer ( grep { $_->type eq "MX" } $packet->answer) {
     $answer->exchange =~ /^(.*)$/;
     dbg("vsender_check: answer: $1");
     push @mxlist, $1;
    }
   }
   push @mxlist, $domain;

   foreach my $vmx (@mxlist) { dbg("vsender_check: mx element: $vmx"); }
   foreach my $mx (@mxlist) {
    dbg("vsender_check: testing mx '$mx' for '$domain'");
    if( $smtp = Net::SMTP->new( $mx, Hello => $HELO, Timeout => 10 ) ){
     dbg("vsender_check: Sent HELO $HELO got banner " .$smtp->banner);
     unless($smtp->mail("<" . $totest . ">")) {
      dbg("vsender_check: mailfrom KO: " .$smtp->message);
      return -1;
     }
     dbg("vsender_check: mailfrom OK: " .$smtp->message);
     unless($smtp->to("<" . $from . ">")) {
      dbg("vsender_check: to KO: " .$smtp->message);
      dbg("vsender_check: to KO: " .$smtp->code);
      # If code starts with 4 add domain to greylisting domains
      if ($smtp->code =~ /^4/) {
       dbg("vsender_check: add to greylisting domains table: " . $domain);
      }
      my $sth = $dbh->prepare( "INSERT OR REPLACE INTO vsendergrey (vdomain) values (?)" );
      $sth->execute( $domain ) or dbg("vsender_check: Database error: " .  $DBI::errstr );
      return 0;
     }
     dbg("vsender_check: to OK: " .$smtp->message);
     my $sth = $dbh->prepare( "INSERT INTO vsender (vfrom, vok) values (?, ?)" );
     $sth->execute( $from, 1) or dbg("vsender_check: Database error: " .  $DBI::errstr );

     $smtp->quit;
     return 1;
    } else {
     dbg("vsender_check: smtp server not listen");
    }
   }
   dbg("vsender_check: End with no positive results from all smtp scanned");
   my $sth = $dbh->prepare( "INSERT INTO vsender (vfrom, vok) values (?, ?)" );
   $sth->execute( $domain, 0) or dbg("vsender_check: Database error: " .  $DBI::errstr );
   return 0;
  }
}

1;
